from typing import Optional
from typing_extensions import Unpack

from loguru import logger

from ..router import CKRouterFunc, CKRouterKwArgs


def log_router_err(error_code: Optional[str] = None):
    def deco(router: CKRouterFunc) -> CKRouterFunc:
        async def wrapper(**kwargs: Unpack[CKRouterKwArgs]):
            try:
                return await router(**kwargs)
            except Exception:
                logger.exception("Error occurred when handling route")
                await kwargs["route"].abort(error_code)

        return wrapper

    return deco



"""Install `cookit[playwright]` before import this module."""

from .router import (
    CKRouterFunc as CKRouterFunc,
    CKRouterInfo as CKRouterInfo,
    CKRouterKwArgs as CKRouterKwArgs,
    CKRouterPattern as CKRouterPattern,
    PWRouter as PWRouter,
    RouterGroup as RouterGroup,
    apply_router_to_page as apply_router_to_page,
    make_real_path_router as make_real_path_router,
)
from .screenshot import (
    WaitForType as WaitForType,
    WaitFunction as WaitFunction,
    screenshot_html as screenshot_html,
    screenshot_selector as screenshot_selector,
)



from typing import Any, Awaitable, Callable, Literal, Optional, Union
from typing_extensions import TypeAlias

from playwright.async_api import Page

WaitFunction: TypeAlias = Callable[[Page], Awaitable[Any]]
WaitForType: TypeAlias = Union[
    Literal["domcontentloaded", "load", "networkidle"],
    WaitFunction,
    None,
]


async def _wait_for(page: Page, wait_type: WaitForType):
    if callable(wait_type):
        await wait_type(page)
    elif wait_type:
        await page.wait_for_load_state(wait_type)


async def screenshot_selector(page: Page, selector: str, **kwargs) -> bytes:
    elem = await page.query_selector(selector)
    if not elem:
        raise ValueError("Element not found")
    return await elem.screenshot(**kwargs)


async def screenshot_html(
    page: Page,
    html: str,
    selector: Optional[str] = None,
    wait_type: WaitForType = "load",
    **kwargs,
) -> bytes:
    await page.set_content(html)
    await _wait_for(page, wait_type)
    if selector:
        return await screenshot_selector(page, selector, **kwargs)
    return await page.screenshot(**kwargs)



import asyncio as aio
from pathlib import Path
from typing import List

import pytest


async def test_with_semaphore():
    from cookit import with_semaphore

    ret = "Hello, World!"
    semaphore = aio.Semaphore(1)

    @with_semaphore(semaphore)
    async def async_func():
        return ret

    result = await aio.gather(
        async_func(),
        async_func(),
    )
    assert tuple(result) == (ret, ret)


async def test_queued():
    from cookit import queued

    counter = 0
    count_list: List[int] = []

    @queued
    async def async_func():
        nonlocal counter
        counter += 1
        await aio.sleep(0)
        count_list.append(counter)

    await aio.gather(
        async_func(),
        async_func(),
        async_func(),
        async_func(),
        async_func(),
    )
    assert count_list == [1, 2, 3, 4, 5]


async def test_race():
    from cookit import race

    async def f1():
        await aio.sleep(1)
        return 1

    async def f2():
        await aio.sleep(2)
        return 2

    async def fe():
        raise Exception

    assert await race(f1(), f2()) == 1
    with pytest.raises(Exception):  # noqa: B017, PT011
        await race(f1(), f2(), fe())


def test_auto_import():
    from cookit import auto_import

    from .auto_import.base import get_counter

    for m in auto_import(
        Path(__file__).parent / "auto_import",
        f"{__package__}.auto_import",
    ):
        if hasattr(m, "main"):
            m.main()
    assert get_counter() == 2

    for m in auto_import(
        Path(__file__).parent / "auto_import",
        f"{__package__}.auto_import",
        lambda _: True,
    ):
        if hasattr(m, "main"):
            m.main()
    assert get_counter() == 5



from playwright.async_api import Error as PWError
from tests.loguru.utils import LoggingContext

from ...utils import expect_exc, mark_test
from ..utils import get_page

ROUTE_BASE_URL = "https://cookit.route"


@mark_test
async def test_log_router_err():
    from cookit.pw import RouterGroup
    from cookit.pw.loguru import log_router_err

    router_group = RouterGroup()

    @router_group.router(f"{ROUTE_BASE_URL}/**/*")
    @log_router_err()
    async def _(**_):
        raise RuntimeError

    async with await get_page() as page:
        await router_group.apply(page)
        with LoggingContext() as ctx:
            with expect_exc(PWError):
                await page.goto(f"{ROUTE_BASE_URL}/")
            ctx.should_log(
                exception=RuntimeError,
                message="Error occurred when handling route",
            )
